import numpy as np
import networkx as nx

from collections import deque

from .constants import log, tol
from .grouping import group, group_rows, boolean_rows
from .geometry import faces_to_edges

try:
    from graph_tool import Graph as GTGraph
    from graph_tool.topology import label_components
    _has_gt = True
except ImportError:
    _has_gt = False
    log.warning('graph-tool unavailable, some operations will be much slower')


def face_adjacency(faces=None, mesh=None, return_edges=False):
    '''
    Returns an (n,2) list of face indices.
    Each pair of faces in the list shares an edge, making them adjacent.


    Arguments
    ----------
    faces: (n, d) int, set of faces referencing vertices by index
    mesh:   Trimesh object, optional if passed will used cached edges
    return_edges: bool, return the edges shared by adjacent faces

    Returns
    ---------
    adjacency: (m,2) int, indexes of faces that are adjacent

    if return_edges:
         edges: (m,2) int, indexes of vertices which make up the
                 edges shared by the adjacent faces

    Example
    ----------
    This is useful for lots of things, such as finding connected components:

    graph = nx.Graph()
    graph.add_edges_from(mesh.face_adjacency)
    groups = nx.connected_components(graph_connected)
    '''

    if mesh is None:
        # first generate the list of edges for the current faces
        # also return the index for which face the edge is from
        edges, edges_face = faces_to_edges(faces, return_index=True)
        edges.sort(axis=1)
    else:
        # if passed a mesh, used the cached values for edges sorted
        edges = mesh.edges_sorted
        edges_face = mesh.edges_face

    # this will return the indices for duplicate edges
    # every edge appears twice in a well constructed mesh
    # so for every row in edge_idx, edges[edge_idx[*][0]] == edges[edge_idx[*][1]]
    # in this call to group rows, we discard edges which don't occur twice
    edge_groups = group_rows(edges, require_count=2)

    if len(edge_groups) == 0:
        log.error('No adjacent faces detected! Did you merge vertices?')

    # the pairs of all adjacent faces
    # so for every row in face_idx, self.faces[face_idx[*][0]] and
    # self.faces[face_idx[*][1]] will share an edge
    face_adjacency = edges_face[edge_groups]
    if return_edges:
        face_adjacency_edges = edges[edge_groups[:, 0]]
        return face_adjacency, face_adjacency_edges
    return face_adjacency


def shared_edges(faces_a, faces_b):
    '''
    Given two sets of faces, find the edges which are in both sets.

    Arguments
    ---------
    faces_a: (n,3) int, set of faces
    faces_b: (m,3) int, set of faces

    Returns
    ---------
    shared: (p, 2) int, set of edges
    '''
    e_a = np.sort(faces_to_edges(faces_a), axis=1)
    e_b = np.sort(faces_to_edges(faces_b), axis=1)
    shared = boolean_rows(e_a, e_b, operation=set.intersection)
    return shared


def connected_edges(G, nodes):
    '''
    Given graph G and list of nodes, return the list of edges that
    are connected to nodes
    '''
    nodes_in_G = deque()
    for node in nodes:
        if not G.has_node(node):
            continue
        nodes_in_G.extend(nx.node_connected_component(G, node))
    edges = G.subgraph(nodes_in_G).edges()
    return edges


def facets(mesh):
    '''
    Find the list of parallel adjacent faces.

    Arguments
    ---------
    mesh:  Trimesh

    Returns
    ---------
    facets: list of groups of face indexes (in mesh.faces) of parallel
            adjacent faces.
    '''
    def facets_nx():
        graph_parallel = nx.from_edgelist(face_idx[parallel])
        facets_idx = np.array([list(i)
                               for i in nx.connected_components(graph_parallel)])
        return facets_idx

    def facets_gt():
        graph_parallel = GTGraph()
        graph_parallel.add_edge_list(face_idx[parallel])
        connected = label_components(graph_parallel, directed=False)[0].a
        facets_idx = group(connected, min_len=2)
        return facets_idx

    # (n,2) list of adjacent face indices
    face_idx = mesh.face_adjacency

    # test adjacent faces for angle
    normal_pairs = mesh.face_normals[[face_idx]]
    normal_dot = (np.sum(normal_pairs[:, 0, :]
                         * normal_pairs[:, 1, :], axis=1) - 1)**2

    # if normals are actually equal, they are parallel with a high degree of
    # confidence
    parallel = normal_dot < tol.zero
    non_parallel = np.logical_not(parallel)

    # saying that two faces are *not* parallel is susceptible to error
    # so we add a radius check which computes the distance between face
    # centroids and divides it by the dot product of the normals
    # this means that small angles between big faces will have a large
    # radius which we can filter out easily.
    # if you don't do this, floating point error on tiny faces can push
    # the normals past a pure angle threshold even though the actual
    # deviation across the face is extremely small.
    center_sq = np.sum(np.diff(mesh.triangles_center[face_idx],
                               axis=1).reshape((-1, 3)) ** 2, axis=1)
    radius_sq = center_sq[non_parallel] / normal_dot[non_parallel]
    parallel[non_parallel] = radius_sq > tol.facet_rsq

    # graph-tool is ~6x faster than networkx but is more difficult to install
    if _has_gt:
        return facets_gt()
    else:
        return facets_nx()


def split(mesh, only_watertight=True, adjacency=None):
    '''
    Given a mesh, will split it up into a list of meshes based on face connectivity
    If only_watertight is true, it will only return meshes where each face has
    exactly 3 adjacent faces.

    Arguments
    ----------
    mesh: Trimesh
    only_watertight: if True, only return watertight components
    adjacency: (n,2) list of face adjacency to override using the plain
               adjacency calculated automatically.

    Returns
    ----------
    meshes: list of Trimesh objects
    '''

    def split_nx():
        adjacency_graph = nx.from_edgelist(adjacency)
        # make sure every face has a node, so single triangles
        # aren't discarded (as they aren't adjacent to anything)
        if not only_watertight:
            # if we are allowing non- watertight result add nodes for every
            # face to make sure single, disconnected triangles are in the graph
            adjacency_graph.add_nodes_from(np.arange(len(mesh.faces)))
        components = nx.connected_components(adjacency_graph)
        result = mesh.submesh(components, only_watertight=only_watertight)
        return result

    def split_gt():
        g = GTGraph()
        if not only_watertight:
            # same as above, for single triangles with no adjacency
            g.add_vertex(len(mesh.faces))
        g.add_edge_list(adjacency)
        component_labels = label_components(g, directed=False)[0].a
        components = group(component_labels)
        result = mesh.submesh(components, only_watertight=only_watertight)
        return result

    if adjacency is None:
        adjacency = mesh.face_adjacency

    if _has_gt:
        return split_gt()
    else:
        return split_nx()


def smoothed(mesh, angle):
    '''
    Return a non- watertight version of the mesh which will
    render nicely with smooth shading.

    Arguments
    ---------
    mesh:  Trimesh object
    angle: float, angle in radians, adjacent faces which have normals
           below this angle will be smoothed.

    Returns
    ---------
    smooth: Trimesh object
    '''
    if len(mesh.face_adjacency) == 0:
        return mesh
    angle_ok = mesh.face_adjacency_angles <= angle
    adjacency = mesh.face_adjacency[angle_ok]
    graph = nx.from_edgelist(adjacency)
    graph.add_nodes_from(np.arange(len(mesh.faces)))
    smooth = mesh.submesh(nx.connected_components(graph),
                          only_watertight=False,
                          append=True)
    return smooth


def is_watertight(edges, return_winding=False):
    '''
    Arguments
    ---------
    edges: (n,2) int, set of vertex indices

    Returns
    ---------
    watertight: boolean, whether every edge is contained by two faces
    '''
    edges_sorted = np.sort(edges, axis=1)
    groups = group_rows(edges_sorted, require_count=2)
    watertight = (len(groups) * 2) == len(edges)
    if return_winding:
        opposing = edges[groups].reshape((-1, 4))[:, 1:3].T
        reversed = np.equal(*opposing).all()
        return watertight, reversed
    return watertight



'''
A basic, slow implementation of ray- triangle queries. 
'''
import numpy as np

from ..constants import tol
from ..grouping import unique_rows
from ..intersections import plane_lines

from .. import util


class RayMeshIntersector:
    '''
    An object to query a mesh for ray intersections.
    Precomputes an r-tree for each triangle on the mesh.
    '''

    def __init__(self, mesh):
        self.mesh = mesh
        self._cache = util.Cache(self.mesh.crc)

    @property
    def tree(self):
        if 'tree' in self._cache:
            return self._cache.get('tree')
        else:
            return self._cache.set('tree',
                                   self.mesh.triangles_tree())

    def intersects_id(self,
                      ray_origins,
                      ray_directions,
                      return_any=False):
        '''
        Find the indexes of triangles the rays intersect

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------
        hits: (n) sequence of triangle indexes which hit the ray
        '''
        hits = rays_triangles_id(triangles=self.mesh.triangles,
                                 ray_origins=ray_origins,
                                 ray_directions=ray_directions,
                                 tree=self.tree,
                                 return_any=return_any)
        return hits

    def intersects_location(self,
                            ray_origins,
                            ray_directions,
                            return_id=False):
        '''
        Return unique cartesian locations where rays hit the mesh.
        If you are counting the number of hits a ray had, this method
        should be used as if only the triangle index is used on- edge hits
        will be counted twice.

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions
        return_id: boolean flag, if True return triangle indexes

        Returns
        ---------
        locations: (n) sequence of (m,3) intersection points
        hits:      (n) list of face ids
        '''
        hits = self.intersects_id(ray_origins=ray_origins,
                                  ray_directions=ray_directions)
        locations = ray_triangle_locations(triangles=self.mesh.triangles,
                                           ray_origins=ray_origins,
                                           ray_directions=ray_directions,
                                           intersections=hits,
                                           tri_normals=self.mesh.face_normals)
        if return_id:
            return locations, hits
        return locations

    def intersects_any_triangle(self,
                                ray_origins,
                                ray_directions):
        '''
        Find out whether the rays in question hit *any* triangle on the mesh.

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------
        hits_any: (n) boolean array of whether or not each ray hit any triangle
        '''
        hits = self.intersects_id(ray_origins, ray_directions)
        hits_any = np.array([len(i) > 0 for i in hits])
        return hits_any

    def intersects_any(self,
                       ray_origins,
                       ray_directions):
        '''
        Find out whether *any* ray hit *any* triangle on the mesh.
        Equivilant to but signifigantly faster than (due to early exit):
            intersects_any_triangle(rays).any()

        Arguments
        ---------
        rays: (n, 2, 3) array of ray origins and directions

        Returns
        ---------
        hit: boolean, whether any ray hit any triangle on the mesh
        '''
        hit = self.intersects_id(ray_origins,
                                 ray_directions,
                                 return_any=True)
        return hit


def rays_triangles_id(triangles,
                      ray_origins,
                      ray_directions,
                      ray_candidates=None,
                      tree=None,
                      return_any=False):
    '''
    Intersect a set of rays and triangles.

    Arguments
    ---------
    triangles:      (n, 3, 3) float array of triangle vertices
    ray_origins:    (m,3) float, array of ray origins
    ray_directions: (m,3) float, array of ray directions
    ray_candidates: (m, *) int array of which triangles are candidates
                    for the ray.
    return_any:     bool, exit loop early if any ray hits any triangle
                    and change output of function to bool

    Returns
    ---------
    if return_any:
        hit:           bool, whether the set of rays hit any triangle
    else:
        intersections: (m) sequence of triangle indexes hit by rays
    '''

    if ray_candidates is None:
        ray_candidates = ray_triangle_candidates(ray_origins=ray_origins,
                                                 ray_directions=ray_directions,
                                                 tree=tree)
    # default set of candidate triangles to be queried
    # is every triangle. this is very slow
    candidates = np.ones(len(triangles), dtype=np.bool)
    hits = [[]] * len(ray_origins)

    for ray_index, ray_ori, ray_dir in zip(range(len(ray_origins)),
                                           ray_origins,
                                           ray_directions):
        # query the triangle candidates
        candidates = ray_candidates[ray_index]
        if len(candidates) == 0: 
            continue
        hit = ray_triangles(triangles[candidates],
                            ray_ori,
                            ray_dir)
        if return_any:
            if hit.any():
                return True
        else:
            hits[ray_index] = np.array(candidates)[hit]

    if return_any:
        return False
    return np.array(hits)


def ray_triangles(triangles,
                  ray_origin,
                  ray_direction):
    '''
    Intersection of multiple triangles and a single ray.

    Moller-Trumbore intersection algorithm.
    '''
    candidates = np.ones(len(triangles), dtype=np.bool)

    # edge vectors and vertex locations in (n,3) format
    vert0 = triangles[:, 0, :]
    vert1 = triangles[:, 1, :]
    vert2 = triangles[:, 2, :]
    edge0 = vert1 - vert0
    edge1 = vert2 - vert0

    # P is a vector perpendicular to the ray direction and one
    # triangle edge.
    P = np.cross(ray_direction, edge1)

    # if determinant is near zero, ray lies in plane of triangle
    det = util.diagonal_dot(edge0, P)
    candidates[np.abs(det) < tol.zero] = False

    if not candidates.any():
        return candidates
    # remove previously calculated terms which are no longer candidates
    inv_det = 1.0 / det[candidates]
    T = ray_origin - vert0[candidates]
    u = util.diagonal_dot(T, P[candidates]) * inv_det

    new_candidates = np.logical_not(np.logical_or(u < -tol.zero,
                                                  u > (1 + tol.zero)))
    candidates[candidates] = new_candidates
    if not candidates.any():
        return candidates
    inv_det = inv_det[new_candidates]
    T = T[new_candidates]
    u = u[new_candidates]

    Q = np.cross(T, edge0[candidates])
    v = np.dot(ray_direction, Q.T) * inv_det

    new_candidates = np.logical_not(np.logical_or((v < -tol.zero),
                                                  (u + v > (1 + tol.zero))))

    candidates[candidates] = new_candidates
    if not candidates.any():
        return candidates

    Q = Q[new_candidates]
    inv_det = inv_det[new_candidates]

    t = util.diagonal_dot(edge1[candidates], Q) * inv_det
    candidates[candidates] = t > tol.zero

    return candidates


def ray_triangle_candidates(ray_origins,
                            ray_directions,
                            tree):
    '''
    Do broad- phase search for triangles that the rays
    may intersect.

    Does this by creating a bounding box for the ray as it
    passes through the volume occupied by the tree
    '''
    ray_bounding = ray_bounds(ray_origins=ray_origins,
                              ray_directions=ray_directions,
                              bounds=tree.bounds)
    ray_candidates = [None] * len(ray_origins)
    for ray_index, bounds in enumerate(ray_bounding):
        ray_candidates[ray_index] = np.array(list(tree.intersection(bounds)))
    return ray_candidates


def ray_bounds(ray_origins,
               ray_directions,
               bounds,
               buffer_dist=1e-5):
    '''
    Given a set of rays and a bounding box for the volume of interest
    where the rays will be passing through, find the bounding boxes
    of the rays as they pass through the volume.

    Arguments
    ---------
    rays: (n,2,3) array of ray origins and directions
    bounds: (2,3) bounding box (min, max)
    buffer_dist: float, distance to pad zero width bounding boxes

    Returns
    ---------
    ray_bounding: (n) set of AABB of rays passing through volume
    '''

    # bounding box we are testing against
    bounds = np.array(bounds)

    # find the primary axis of the vector
    axis = np.abs(ray_directions).argmax(axis=1)
    axis_bound = bounds.reshape((2, -1)).T[axis]
    axis_ori = np.array([ray_origins[i][a]
                         for i, a in enumerate(axis)]).reshape((-1, 1))
    axis_dir = np.array([ray_directions[i][a]
                         for i, a in enumerate(axis)]).reshape((-1, 1))

    # parametric equation of a line
    # point = direction*t + origin
    # p = dt + o
    # t = (p-o)/d
    t = (axis_bound - axis_ori) / axis_dir

    # prevent the bounding box from including triangles
    # behind the ray origin
    t[t < buffer_dist] = buffer_dist

    # the value of t for both the upper and lower bounds
    t_a = t[:, 0].reshape((-1, 1))
    t_b = t[:, 1].reshape((-1, 1))

    # the cartesion point for where the line hits the plane defined by
    # axis
    on_a = (ray_directions * t_a) + ray_origins
    on_b = (ray_directions * t_b) + ray_origins

    on_plane = np.column_stack(
        (on_a, on_b)).reshape(
        (-1, 2, ray_directions.shape[1]))

    ray_bounding = np.hstack((on_plane.min(axis=1), on_plane.max(axis=1)))
    # pad the bounding box by TOL_BUFFER
    # not sure if this is necessary, but if the ray is  axis aligned
    # this function will otherwise return zero volume bounding boxes
    # which may or may not screw up the r-tree intersection queries
    ray_bounding += np.array([-1, -1, -1, 1, 1, 1]) * buffer_dist

    return ray_bounding


def ray_triangle_locations(triangles,
                           ray_origins,
                           ray_directions,
                           intersections,
                           tri_normals):
    '''
    Given a set of triangles, rays, and intersections between the two,
    find the cartesian locations of the intersections points.

    Arguments
    ----------
    triangles:     (n, 3, 3) set of triangle vertices
    rays:          (m, 2, 3) set of ray origins/ray direction pairs
    intersections: (m) sequence of intersection indidices which triangles
                    each ray hits.

    Returns
    ----------
    locations: (m) sequence of (p,3) cartesian points
    '''

    ray_segments = np.array([ray_origins,
                             ray_origins + ray_directions])
    locations = [[]] * len(ray_origins)

    for r, tri_group in enumerate(intersections):
        group_locations = np.zeros((len(tri_group), 3))
        valid = np.zeros(len(tri_group), dtype=np.bool)
        for i, tri_index in enumerate(tri_group):
            origin = triangles[tri_index][0]
            normal = tri_normals[tri_index]
            segment = ray_segments[:, r, :].reshape((2, -1, 3))
            point, ok = plane_lines(plane_origin=origin,
                                    plane_normal=normal,
                                    endpoints=segment,
                                    line_segments=False)
            if ok:
                valid[i] = True
                group_locations[i] = point
        group_locations = group_locations[valid]
        unique = unique_rows(group_locations)[0]
        locations[r] = group_locations[unique]
    return np.array(locations)



import numpy as np

from ..grouping import group_rows
from ..transformations import rotation_matrix, transform_points
from .transforms import TransformForest

from .. import util


from collections import deque, OrderedDict


class Scene:
    '''
    A simple scene graph which can be rendered directly via pyglet/openGL,
    or through other endpoints such as a raytracer.

    Meshes and lights are added by name, which can then be moved by updating
    transform in the transform tree.
    '''

    def __init__(self,
                 node=None,
                 base_frame='world'):

        # instance name : mesh name
        self.nodes = {}
        self._cache = util.Cache(id_function=self.md5)

        # mesh name : Trimesh object
        self.geometry = OrderedDict()
        self.flags = {}
        self.transforms = TransformForest(base_frame=base_frame)

        if node is not None:
            self.add_geometry(node)
            self.set_camera()
        self._cache.id_set()

    def add_geometry(self, geometry):
        '''
        Add a geometry to the scene.

        If the mesh has multiple transforms defined in its metadata, they will
        all be copied into the TransformForest of the current scene automatically.

        Arguments
        ----------
        geometry: Trimesh, Path3D, or list of same
        '''
        if util.is_sequence(geometry):
            return [self.add_geometry(i) for i in geometry]

        if 'name' in geometry.metadata:
            name_geometry = geometry.metadata['name']
        else:
            name_geometry = 'geometry_' + str(len(self.geometry))

        self.geometry[name_geometry] = geometry

        if 'transforms' in geometry.metadata:
            transforms = np.array(
                geometry.metadata['transforms']).reshape((-1, 4, 4))
        else:
            transforms = np.eye(4).reshape((-1, 4, 4))

        for i, transform in enumerate(transforms):
            name_node = name_geometry + '_' + str(i)
            self.nodes[name_node] = name_geometry
            self.flags[name_node] = {'visible': True}
            self.transforms.update(frame_to=name_node,
                                   matrix=transform)

    def md5(self):
        '''
        MD5 of scene, which will change when meshes or transforms are changed
        '''
        mesh_hash = util.md5_object(
            np.sort([hash(i) for i in self.geometry.values()]))
        result = mesh_hash + self.transforms.md5()
        return result

    @util.cache_decorator
    def bounds(self):
        '''
        Return the overall bounding box of the scene.

        Returns
        --------
        bounds: (2,3) float points for min, max corner
        '''
        corners = deque()
        for instance, mesh_name in self.nodes.items():
            transform = self.transforms.get(instance)
            corners.append(transform_points(self.geometry[mesh_name].bounds,
                                            transform))
        corners = np.vstack(corners)
        bounds = np.array([corners.min(axis=0),
                           corners.max(axis=0)])
        return bounds

    @util.cache_decorator
    def extents(self):
        return np.diff(self.bounds, axis=0).reshape(-1)

    @util.cache_decorator
    def scale(self):
        '''
        The approximate scale of the mesh

        Returns
        -----------
        scale: float, the mean of the bounding box edge lengths
        '''
        return self.extents.mean()

    @util.cache_decorator
    def centroid(self):
        '''
        Return the center of the bounding box for the scene.

        Returns
        --------
        centroid: (3) float point for center of bounding box
        '''
        centroid = np.mean(self.bounds, axis=0)
        return centroid

    @util.cache_decorator
    def triangles(self):
        '''
        Return a correctly transformed polygon soup of the current scene.

        Returns
        ----------
        triangles: (n,3,3) float, triangles in space
        '''
        triangles = deque()
        triangles_index = deque()
        for index, node_info in zip(range(len(self.nodes)),
                                    self.nodes.items()):
            node, geometry_name = node_info
            geometry = self.geometry[geometry_name]
            if not hasattr(geometry, 'triangles'):
                continue
            transform = self.transforms.get(node)
            triangles.append(transform_points(geometry.triangles.reshape((-1,3)),
                                              transform))
            triangles_index.append(np.ones(len(geometry.triangles),
                                           dtype=np.int64) * index)
        triangles = np.vstack(triangles).reshape((-1,3,3))
        self._cache['triangles_index'] = np.hstack(triangles_index)
        return triangles

    @util.cache_decorator
    def triangles_index(self):
        '''
        Which index of self.nodes.values() does each triangle come from.

        Returns
        ---------
        triangles_index: (len(self.triangles),) int, index of self.nodes.values()
        '''
        populate = self.triangles
        return self._cache['triangles_index']
    
    def duplicate_nodes(self):
        '''
        Return a sequence of node keys of identical meshes.

        Will combine meshes duplicated by copying in space with different keys in
        self.geometry, as well as meshes repeated by self.nodes.

        Returns
        -----------
        duplicates: (m) sequence of keys to self.nodes that represent identical geometry
        '''
        mesh_ids = {k: m.identifier for k, m in self.geometry.items()}

        node_keys = np.array(list(self.nodes.keys()))
        node_ids = [mesh_ids[v] for v in self.nodes.values()]

        node_groups = group_rows(node_ids, digits=1)

        duplicates = [np.sort(node_keys[g]).tolist() for g in node_groups]
        return duplicates

    def set_camera(self, angles=None, distance=None, center=None):
        if center is None:
            center = self.centroid
        if distance is None:
            distance = self.extents.max()
        if angles is None:
            angles = np.zeros(3)

        translation = np.eye(4)
        translation[0:3, 3] = center
        translation[2][3] += distance * 1.5

        transform = np.dot(rotation_matrix(angles[0], [1, 0, 0], point=center),
                           rotation_matrix(angles[1], [0, 1, 0], point=center))
        transform = np.dot(transform, translation)

        self.transforms.update(frame_from='camera',
                               frame_to=self.transforms.base_frame,
                               matrix=transform)

    def dump(self):
        '''
        Append all meshes in scene to a list of meshes.
        '''
        result = deque()
        for node_id, mesh_id in self.nodes.items():
            transform = self.transforms.get(node_id)
            current = self.geometry[mesh_id].copy()
            current.apply_transform(transform)
            result.append(current)
        return np.array(result)

    def export(self, file_type='dict64'):
        '''
        Export a snapshot of the current scene.

        Arguments
        ----------
        file_type: what encoding to use for meshes
                   ie: dict, dict64, stl

        Returns
        ----------
        export: dict with keys:
                meshes: list of meshes, encoded as per file_type
                transforms: edge list of transforms, eg:
                             ((u, v, {'matrix' : np.eye(4)}))
        '''
        export = {'transforms': self.transforms.export(),
                  'nodes': self.nodes,
                  'meshes': {},
                  'scene_cache': self._cache.cache}
        # if the mesh has an export method use it, otherwise put the mesh
        # itself into the export object
        for node, mesh in self.geometry.items():
            if hasattr(mesh, 'export'):
                export['meshes'][node] = mesh.export(file_type=file_type)
            else:
                export['meshes'][node] = mesh
        return export

    def save_image(self, file_obj, resolution=(1024, 768), **kwargs):
        from .viewer import SceneViewer
        SceneViewer(self,
                    save_image=file_obj,
                    resolution=resolution,
                    **kwargs)

    def explode(self, vector=[0.0, 0.0, 1.0], origin=None):
        '''
        Explode a scene around a point and vector.
        '''
        if origin is None:
            origin = self.centroid
        centroids = np.array(
            [self.geometry[i].centroid for i in self.nodes.values()])

        if np.shape(vector) == (3,):
            vectors = np.tile(vector, (len(centroids), 1))
            projected = np.dot(vector, (centroids - origin).T)
            offsets = vectors * projected.reshape((-1, 1))
        elif isinstance(vector, float):
            offsets = (centroids - origin) * vector
        else:
            raise ValueError('Explode vector must by (3,) or float')

        for offset, node_key in zip(offsets, self.nodes.keys()):
            current = self.transforms[node_key]
            current[0:3, 3] += offset
            self.transforms[node_key] = current

    def show(self, block=True, **kwargs):
        # this imports pyglet, and will raise an ImportError
        # if pyglet is not available
        from .viewer import SceneViewer

        def viewer():
            SceneViewer(self, **kwargs)

        if block:
            viewer()
        else:
            from threading import Thread
            Thread(target=viewer, kwargs=kwargs).start()



import numpy as np
import traceback
import os

from .. import util

from ..base import Trimesh
from ..constants import _log_time, log
from ..util import is_file, is_string, make_sequence, is_instance_named

from .assimp import _assimp_loaders
from .stl import _stl_loaders
from .misc import _misc_loaders
from .step import _step_loaders
from .ply import _ply_loaders

try:
    from ..path.io.load import load_path, path_formats
except:
    _path_traceback = traceback.format_exc(4)

    def load_path(*args, **kwargs):
        '''
        Dummy load path function that will raise an exception on use.
        Import of path failed, probably because a dependency is not installed.
        '''
        print(_path_traceback)
        raise ImportError('No path functionality available!')

    def path_formats():
        return []


def mesh_formats():
    return list(_mesh_loaders.keys())


def available_formats():
    return np.append(mesh_formats(), path_formats())


def load(file_obj, file_type=None, **kwargs):
    '''
    Load a mesh or vectorized path into a Trimesh, Path2D, or Path3D object.

    Arguments
    ---------
    file_obj: a filename string or a file-like object
    file_type: str representing file type (eg: 'stl')

    Returns
    ---------
    geometry: Trimesh, Path2D, Path3D, or list of same.
    '''
    # check to see if we're trying to load something that is already a Trimesh
    out_types = ('Trimesh', 'Path')
    if any(is_instance_named(file_obj, t) for t in out_types):
        log.info('Loaded called on %s object, returning input',
                 file_obj.__class__.__name__)
        return file_obj

    file_obj, file_type, metadata = _parse_file_args(file_obj, file_type)

    if file_type in path_formats():
        loaded = load_path(file_obj, file_type, **kwargs)
    elif file_type in mesh_formats():
        loaded = load_mesh(file_obj, file_type, **kwargs)
    else:
        raise ValueError('File type: %s not supported', str(file_type))

    for i in util.make_sequence(loaded):
        # check to make sure loader actually loaded something
        # assert any(is_instance_named(i, t) for t in out_types)
        i.metadata.update(metadata)

    return loaded


@_log_time
def load_mesh(file_obj, file_type=None):
    '''
    Load a mesh file into a Trimesh object

    Arguments
    ---------
    file_obj: a filename string or a file-like object
    file_type: str representing file type (eg: 'stl')

    Returns:
    ----------
    mesh: a single Trimesh object, or a list of Trimesh objects,
          depending on the file format.

    '''
    file_obj, file_type, metadata = _parse_file_args(file_obj, file_type)
    loaded = _mesh_loaders[file_type](file_obj,
                                      file_type)
    if is_file(file_obj):
        file_obj.close()

    log.debug('loaded mesh using %s',
              _mesh_loaders[file_type].__name__)

    meshes = [Trimesh(**i) for i in make_sequence(loaded)]
    for i in meshes:
        i.metadata.update(metadata)

    if len(meshes) == 1:
        return meshes[0]
    return meshes


def _parse_file_args(file_obj, file_type):
    '''
    Given a file_obj and a file_type, try to turn them into a file-like object
    and a lowercase string of file type

    Arguments
    -----------
    file_obj:  str: if string represents a file path, returns
                    -------------------------------------------
                    file_obj:   an 'rb' opened file object of the path
                    file_type:  the extension from the file path

               str: if string is NOT a path, but has JSON-like special characters
                    -------------------------------------------
                    file_obj:   the same string passed as file_obj
                    file_type:  set to 'json'

               str: string is not an existing path or a JSON-like object
                    -------------------------------------------
                    ValueError will be raised as we can't do anything with input

               file like object: we cannot grab information on file_type automatically
                    -------------------------------------------
                    ValueError will be raised if file_type is None
                    file_obj:  same as input
                    file_type: same as input

               other object: like a shapely.geometry.Polygon, etc:
                    -------------------------------------------
                    file_obj:  same as input
                    file_type: if None initially, set to the class name
                               (in lower case), otherwise passed through

    file_type: str, type of file and handled according to above

    Returns
    -----------
    file_obj:  loadable object
    file_type: str, lower case of the type of file (eg 'stl', 'dae', etc)
    '''
    metadata = {}

    if util.is_file(file_obj) and file_type is None:
        raise ValueError(
            'File type must be specified when passing file objects!')
    if util.is_string(file_obj):
        try:
            exists = os.path.isfile(file_obj)
        except:
            exists = False
        if exists:
            metadata['file_path'] = file_obj
            metadata['file_name'] = os.path.basename(file_obj)
            # if file_obj is a path that exists use extension as file_type
            file_type = (str(file_obj).split('.')[-1])
            file_obj = open(file_obj, 'rb')
        else:
            if file_type is not None:
                return file_obj, file_type, metadata
            elif '{' in file_obj:
                # if a dict bracket is in the string, its probably a straight
                # JSON
                file_type = 'json'
            else:
                raise ValueError(
                    'File object passed as string that is not a file!')

    if file_type is None:
        file_type = file_obj.__class__.__name__

    if is_string(file_type) and '.' in file_type:
        # if someone has passed the whole filename as the file_type
        # use the file extension as the file_type
        metadata['file_name'] = os.path.basename(file_type)
        file_type = file_type.split('.')[-1]
    file_type = file_type.lower()
    return file_obj, file_type, metadata

_mesh_loaders = {}
# assimp has a lot of loaders, but they are all quite slow
# so we load them first and replace them with native loaders if possible
_mesh_loaders.update(_assimp_loaders)
_mesh_loaders.update(_stl_loaders)
_mesh_loaders.update(_misc_loaders)
_mesh_loaders.update(_step_loaders)
_mesh_loaders.update(_ply_loaders)
